package org.fictus

trait Assertions {
    implicit def anyToAssertion[A](value: A) = Assertion(value)
    implicit def seqToAssertion[A](value: Seq[A]) = SeqAssertion(value)
    implicit def stringToAssertion[A](value: String) = StringAssertion(value)

    def fail(): Nothing = fail("")
    def fail(msg: => String): Nothing = throw new FailedError(msg)
}

case class SeqAssertion[A](actual: Seq[A]) {
    def mustExist[B <: A](predicate: A => Boolean): SeqAssertion[A] = {
        if (!actual.exists(predicate)) throw new AssertionError("Expected " + actual + " to adhere to predicate, but didn't.")
        else this
    }

    def mustContain[B <: A](expected: B): SeqAssertion[A] =
        if (!actual.contains(expected)) throw new AssertionError("Expected " + actual + " to contain " + expected)
        else this

    def mustContain[B <: A](expected: B*): SeqAssertion[A] = {
        import scala.collection.mutable.ListBuffer
        val contained = new ListBuffer[B]
        val missing = new ListBuffer[B]
        expected.foreach { value =>
            if (actual.contains(value)) contained += value
            else missing += value
        }
        if (!missing.isEmpty)
            throw new AssertionError("Expected " + actual + " to contain (" + expected.mkString(", ") + "), but was missing: " + missing.mkString(", "))
        else this
    }

    override def equals(arg: Any) = throw new PebkacError("Using equals will lead to a buggy test... use equalTo instead")
}

case class StringAssertion(actual: String) {
    def mustContain(expected: String): StringAssertion = actual match {
        case s: String if (s.contains(expected)) => this
        case actual                              => throw new AssertionError("Expected " + actual + " to contain " + expected)
    }

    def mustStartWith(expected: String): StringAssertion = actual match {
        case s: String if (s.startsWith(expected)) => this
        case actual                                => throw new AssertionError("Expected " + actual + " to start with " + expected)
    }

    def mustEndWith(expected: String): StringAssertion = actual match {
        case s: String if (s.endsWith(expected)) => this
        case actual                              => throw new AssertionError("Expected " + actual + " to end with " + expected)
    }

    def mustEqualIgnoringCase(expected: String): StringAssertion = actual match {
        case s: String if (s.equalsIgnoreCase(expected)) => this
        case unexpected                                  => throw new AssertionError("Expected\n" + expected + " but was\n" + unexpected)
    }

    override def equals(arg: Any) = throw new PebkacError("Using equals will lead to a buggy test... use equalTo instead")
}

case class Assertion[A](actual: A) {
    def mustMatch[B](success: PartialFunction[A, B]) = {
        if (success.isDefinedAt(actual)) success(actual)
        else throw new MatcherError("Unexpected value:\n" + actual + "\ndid not match your case!")
        this
    }

    def mustNotMatch[B](fail: PartialFunction[A, B]) = {
        if (fail.isDefinedAt(actual)) throw new MatcherError("Unexpected value:\n" + actual + "\nmatched your case unexpectedly!")
        else this
    }

    def mustBe[B <: AnyRef](expected: B): Assertion[A] = actual match {
        case good: AnyRef if (expected eq good) => this
        case _                                  => throw new AssertionError("Expected\n" + expected + " but was\n" + actual)
    }

    def mustEqual[B <: A](expected: B): Assertion[A] = {
        (actual == expected) match {
            case true  => this
            case false => throw new AssertionError("Expected\n" + expected + " but was\n" + actual)
        }
    }

    def mustNotEqual[B <: A](expected: B): Assertion[A] = {
        (actual == expected) match {
            case true  => throw new AssertionError("Expected\n" + expected + " != " + actual)
            case false => this
        }
    }

    override def equals(arg: Any) = throw new PebkacError("Using equals will lead to a buggy test... use equalTo instead")
}